/************************************************************************
 *   IRC - Internet Relay Chat, modules/m_servid.c
 *   Copyright (C) 2001 Hybrid Development Team
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: m_servid.c,v 1.0 2003/01/02 22:15:33 UniBG Exp $
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "ircd.h"
#include "config.h"
#include "irc_string.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "channel.h"
#include "hash.h"

#ifdef UniBG
static void ms_servid(struct Client *, struct Client *, int, char **);
static void ns_parse(struct Client *, struct Client *, int, char **);
static void os_parse(struct Client *, struct Client *, int, char **);
static void nk_parse(struct Client *, struct Client *, int, char **);

struct ServIDStruct
{
  char *cmd;
  void (*handler)();
};

static struct ServIDStruct servid_cmd_table[] =
{
  { "NS"      , ns_parse,       },
  { "OS"      , os_parse,       },
  { "NK"      , nk_parse,       },
  { (char *) 0, (void (*)()) 0, }
};

struct Message servid_msgtab = {
  "SERVID", 0, 0, 1, 1, MFLG_SLOW, 0,
  { m_ignore, m_ignore, ms_servid, m_ignore, m_ignore }
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
  mod_add_cmd(&servid_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&servid_msgtab);
}

const char *_version = "$Revision: 1.0 $";
#endif

/* ms_servid
 *
 * inputs:
 *   parv[0] = sender prefix
 *   parv[1] = NS | OS | NK
 *   parv[2] = nickname
 *   parv[3] = 1  | 0
 *   parv[4] = TS (not required)
 *
 * As we _SHOULD_ notice, the SERVID is missing here, it's not
 * passed by to ms_servid, which is a good think.
 *
 * We may miss the IsServer check, while we are sure it's a server!
 *
 * There are some funky sh!ts in the ircd-hybrid-6.0+UniBG-p2 m_servid()...
 * As from now they're gone ;)
 * Some more checkups are made (needed and unneed)
 *
 * The CAP_SERVID was a good idea... 
 * Changing IsClient to IsPerson too...
 * Getting ServicesOn check and EmptryString too...
 *
 ** Main code taken from fuis (but reworked).
 */
static void ms_servid(struct Client *client_p, struct Client *source_p,
                   int parc, char *parv[])
{
  int i;

  if (parc < 4 || EmptyString(parv[3]) || !ServicesOn)
    return;

  for (i = 0; servid_cmd_table[i].handler; i++)
    if (!irccmp(servid_cmd_table[i].cmd, parv[1]))
      servid_cmd_table[i].handler(client_p, source_p, parc, parv);
}    

/* ns_parse
 *
 * inputs:
 *   parv[0] = sender prefix
 *   parv[1] = NS
 *   parv[2] = nickname
 *   parv[3] = 1 | 0
 *   parv[4] = TS (not required)
 */
static void ns_parse(struct Client *client_p, struct Client *source_p,
                   int parc, char *parv[])
{
  struct Client *target_p;

  if ((target_p = find_person(parv[2])) == NULL)
    return;

  if (!IsPerson(target_p))
    return;

  if (parv[3][0] == '1')
  {
    if (!IsNSId(target_p))
    {
      CurrentIdentUsers++;

      if (CurrentIdentUsers > MaxIdentUsers)
        MaxIdentUsers = CurrentIdentUsers;
    }

    SetNSId(target_p);
  }
  else if (parv[3][0] == '0')
  {
    if (IsNSId(target_p))
      CurrentIdentUsers--;

    ClearNSId(target_p);
  }

  if (parc == 5 && !EmptyString(parv[4]))
    target_p->ns_reg = (long) atoi(parv[4]);
  else
    target_p->ns_reg = 0;

  sendto_server(source_p, target_p, NULL, CAP_SERVID, NOCAPS, NOFLAGS,
                "SERVID NS %s %c %s",
                target_p->name, parv[3][0],
                parc == 5 ? parv[4] : "");
}

/* os_parse
 *
 * inputs:
 *   parv[0] = sender prefix
 *   parv[1] = OS
 *   parv[2] = nickname
 *   parv[3] = 1 | 0
 */
static void os_parse(struct Client *client_p, struct Client *source_p,
                   int parc, char *parv[])
{
  struct Client *target_p;

  if ((target_p = find_person(parv[2])) == NULL)
    return;

  if (!IsPerson(target_p))
    return;

  if (parv[3][0] == '1')
    SetOSId(target_p);
  else if (parv[3][0] == '0')
    ClearOSId(target_p);

  sendto_server(source_p, target_p, NULL, CAP_SERVID, NOCAPS, NOFLAGS,
                "SERVID OS %s %c",
                target_p->name, parv[3][0]);
}

/* nk_parse
 *
 * inputs:
 *   parv[0] = sender prefix
 *   parv[1] = NK
 *   parv[2] = nickname
 *   parv[3] = 1 | 0
 */
static void nk_parse(struct Client *client_p, struct Client *source_p,
                   int parc, char *parv[])
{
  struct Client *target_p;

  if ((target_p = find_person(parv[2])) == NULL)
    return;

  if (!IsPerson(target_p))
    return;

  if (parv[3][0] == '1')
    SetNetKeeper(target_p);
  else if (parv[3][0] == '0')
    ClearNetKeeper(target_p);

  sendto_server(source_p, target_p, NULL, CAP_SERVID, NOCAPS, NOFLAGS,
                "SERVID NK %s %c",
                target_p->name, parv[3][0]);
}
#endif
